import logging
import os
import sys
from datetime import datetime
from glob import glob
from configparser import ConfigParser

import joblib

from ..utils.data import make_spects_from_list_of_files, make_data_dicts
from ..utils.mat import convert_mat_to_spect
from .. import config


def make_data(labelset,
              all_labels_are_int,
              data_dir,
              total_train_set_dur,
              val_dur,
              test_dur,
              config_file,
              silent_gap_label=0,
              skip_files_with_labels_not_in_labelset=True,
              output_dir=None,
              mat_spect_files_path=None,
              mat_spects_annotation_file=None,
              spect_params=None,
              ):
    """make datasets for training models to generate a learning curve

    Parameters
    ----------
    labelset : list
        of str or int, set of labels for syllables
    all_labels_are_int : bool
        if True, labels are of type int, not str
    data_dir : str
        path to directory with audio files from which to make dataset
    total_train_set_dur : float
        total duration of training set, in seconds.
        Training subsets of shorter duration will be drawn from this set.
    val_dur : float
        total duration of validation set, in seconds.
    test_dur : float
        total duration of test set, in seconds.
    silent_gap_label : int
        label for time bins of silent gaps between syllables.
        Type is int because labels are converted to a set of
        n consecutive integers {0,1,2...n} where n is the number
        of syllable classes + the silent gap class.
        Default is 0 (in which case labels are {1,2,3,...,n}).
    skip_files_with_labels_not_in_labelset : bool
        if True, skip a file if the labels variable contains labels not
        found in 'labelset'. Default is True.
    output_dir : str
        Path to location where data sets should be saved. Default is None,
        in which case data sets are saved in the current working directory.
    mat_spect_files_path : str
        Path to a directory of .mat files that contain spectrograms.
        Default is None (and this implies user is supplying audio files
         instead of supplying spectrograms in .mat files).
    mat_spects_annotation_file : str
        Path to annotation file associated with .mat files.
        Default is None.
    spect_params : dict
        Dictionary of parameters for creating spectrograms.
        Default is None (implying that spectrograms are already made).

    Returns
    -------
    None

    Saves .spect files and data_dicts in output_dir specified by user.
    """
    logger = logging.getLogger(__name__)
    logger.setLevel('INFO')

    # map labels to series of consecutive integers from 0 to n inclusive
    # where 0 is the label for silent periods between syllables
    # and n is the number of syllable labels
    if all_labels_are_int:
        labels_mapping = dict(zip([int(label) for label in labelset],
                                  range(1, len(labelset) + 1)))
    else:
        labels_mapping = dict(zip(labelset,
                                  range(1, len(labelset) + 1)))
    labels_mapping['silent_gap_label'] = silent_gap_label
    if sorted(labels_mapping.values()) != list(range(len(labels_mapping))):
        raise ValueError('Labels mapping does not map to a consecutive'
                         'series of integers from 0 to n (where 0 is the '
                         'silent gap label and n is the number of syllable'
                         'labels).')

    timenow = datetime.now().strftime('%y%m%d_%H%M%S')

    if mat_spect_files_path:
        print('will use spectrograms from .mat files in {}'
              .format(mat_spect_files_path))
        mat_spect_files = glob(os.path.join(mat_spect_files_path, '*.mat'))
        if output_dir is None:
            output_dir = os.path.join(mat_spect_files_path,
                                  'spectrograms_' + timenow)
        os.mkdir(output_dir)

        spect_files_path = convert_mat_to_spect(mat_spect_files,
                                                mat_spects_annotation_file,
                                                output_dir,
                                                labels_mapping=labels_mapping)

    else:
        logger.info('will make training data from: {}'.format(data_dir))
        if output_dir is None:
            output_dir = os.path.join(data_dir,
                                  'spectrograms_' + timenow)
        else:
            output_dir = os.path.join(output_dir,
                                  'spectrograms_' + timenow)
        os.mkdir(output_dir)

        cbins = glob(os.path.join(data_dir, '*.cbin'))
        if cbins == []:
            # if we don't find .cbins in data_dir, look in sub-directories
            cbins = []
            subdirs = glob(os.path.join(data_dir,'*/'))
            for subdir in subdirs:
                cbins.extend(glob(os.path.join(data_dir,
                                               subdir,
                                               '*.cbin')))
        if cbins == []:
            # try looking for .wav files
            wavs = glob(os.path.join(data_dir, '*.wav'))

            if cbins == [] and wavs == []:
                raise FileNotFoundError('No .cbin or .wav files found in {} or'
                                        'immediate sub-directories'
                                        .format(data_dir))
            # look for canary annotation
            annotation_file = glob(os.path.join(data_dir, '*annotation*.mat'))
            if len(annotation_file) == 1:
                annotation_file = annotation_file[0]
            else:  # try Koumura song annotation
                annotation_file = glob(os.path.join(data_dir, '../Annotation.xml'))
                if len(annotation_file) == 1:
                    annotation_file = annotation_file[0]
                else:
                    raise ValueError('Found more than one annotation.mat file: {}. '
                                     'Please include only one such file in the directory.'
                                     .format(annotation_file))

        if cbins:
            spect_files_path = \
                make_spects_from_list_of_files(cbins,
                                               spect_params,
                                               output_dir,
                                               labels_mapping,
                                               skip_files_with_labels_not_in_labelset)
        elif wavs:
            spect_files_path = \
                make_spects_from_list_of_files(wavs,
                                               spect_params,
                                               output_dir,
                                               labels_mapping,
                                               skip_files_with_labels_not_in_labelset,
                                               annotation_file)

    saved_data_dict_paths = make_data_dicts(output_dir,
                                            total_train_set_dur,
                                            val_dur,
                                            test_dur,
                                            labelset,
                                            spect_files_path)

    # lastly rewrite config file,
    # so that paths where results were saved are automatically in config.
    config = ConfigParser()
    config.read(config_file)
    for key, saved_data_dict_path in saved_data_dict_paths.items():
        config.set(section='TRAIN',
                   option=key + '_data_path',
                   value=saved_data_dict_path)

    with open(config_file, 'w') as config_file_rewrite:
        config.write(config_file_rewrite)


if __name__ == "__main__":
    config_file = os.path.normpath(sys.argv[1])
    config = config.parse_config(config_file)
    make_data(labelset=config.data.labelset,
              all_labels_are_int=config.data.all_labels_are_int,
              data_dir=config.data.data_dir,
              total_train_set_dur=config.data.total_train_set_dur,
              val_dur=config.data.val_dur,
              test_dur=config.data.test_dur,
              config_file=config_file,
              silent_gap_label=config.data.silent_gap_label,
              skip_files_with_labels_not_in_labelset=config.data.skip_files_with_labels_not_in_labelset,
              output_dir=config.data.output_dir,
              mat_spect_files_path=config.data.mat_spect_files_path,
              mat_spects_annotation_file=config.data.mat_spects_annotation_file,
              spect_params=config.spect_params)



"""parses [TRAIN] section of config"""
from configparser import NoOptionError

import attr
from attr.validators import instance_of, optional

from .validators import is_a_directory, is_a_file
from .. import network


@attr.s
class TrainConfig:
    """class that represents [TRAIN] section of config.ini file

    Attributes
    ----------
    networks : namedtuple
        where each field is the Config tuple for a neural network and the name
        of that field is the name of the class that represents the network.
    train_data_dict_path : str
        path to training data
    num_epochs : int
        number of training epochs. One epoch = one iteration through the entire
        training set.
    normalize_spectrograms : bool
        if True, use spect.utils.data.SpectScaler to normalize the spectrograms.
        Normalization is done by subtracting off the mean for each frequency bin
        of the training set and then dividing by the std for that frequency bin.
        This same normalization is then applied to validation + test data.
    val_data_dict_path : str
        path to validation data. Default is None, in which case accuracy is not measured
        on a validation set during training.
    val_error_step : int
        step/epoch at which to estimate accuracy using validation set.
        Default is None, in which case no validation is done.
    checkpoint_step : int
        step/epoch at which to save to checkpoint file.
        Default is None, in which case checkpoint is only saved at the last epoch.
    patience : int
        number of epochs to wait without the error dropping before stopping the
        training. Default is None, in which case training continues for num_epochs
    train_set_durs : list
        of int, durations in seconds of subsets taken from training data
        to create a learning curve, e.g. [5, 10, 15, 20]. Default is None
        (when training a single model on all available training data).
    num_replicates : int
        number of times to replicate training for each training set duration
        to better estimate mean accuracy for a training set of that size.
        Each replicate uses a different randomly drawn subset of the training
        data (but of the same duration).
    save_only_single_checkpoint_file : bool
        if True, save only one checkpoint file instead of separate files every time
        we save. Default is True.
    use_train_subsets_from_previous_run : bool
        if True, use training subsets saved in a previous run. Default is False.
        Requires setting previous_run_path option in config.ini file.
    previous_run_path : str
        path to results directory from a previous run.
        Used for training if use_train_subsets_from_previous_run is True.
    save_transformed_data : bool
        if True, save transformed data (i.e. scaled, reshaped). The data can then
        be used on a subsequent run of learncurve (e.g. if you want to compare results
        from different hyperparameters across the exact same training set).
        Also useful if you need to check what the data looks like when fed to networks.
    """
    # required for both train and learncurve
    networks = attr.ib(validator=instance_of(list))
    train_data_dict_path = attr.ib(validator=[instance_of(str), is_a_file])
    num_epochs = attr.ib(validator=instance_of(int))

    # used for both train and learncurve, but optional
    normalize_spectrograms = attr.ib(validator=optional(instance_of(bool)), default=False)
    val_data_dict_path = attr.ib(validator=optional([instance_of(str), is_a_file]), default=None)
    test_data_dict_path = attr.ib(validator=optional([instance_of(str), is_a_file]), default=None)
    val_error_step = attr.ib(validator=optional(instance_of(int)), default=None)
    checkpoint_step = attr.ib(validator=optional(instance_of(int)), default=None)
    patience = attr.ib(validator=optional(instance_of(int)), default=None)

    # used for learncurve, not train
    train_set_durs = attr.ib(validator=optional(instance_of(list)), default=None)
    num_replicates = attr.ib(validator=optional(instance_of(int)), default=None)

    # things most users probably won't care about
    save_only_single_checkpoint_file = attr.ib(validator=instance_of(bool), default=True)
    use_train_subsets_from_previous_run = attr.ib(validator=instance_of(bool), default=False)
    previous_run_path = attr.ib(validator=optional([instance_of(str), is_a_directory]), default=None)
    save_transformed_data = attr.ib(validator=instance_of(bool), default=False)


def parse_train_config(config, config_file):
    """parse [TRAIN] section of config.ini file

    Parameters
    ----------
    config : ConfigParser
        containing config.ini file already loaded by parse function

    Returns
    -------
    train_config : vak.config.train.TrainConfig
        instance of TrainConfig class
    """
    config_dict = {}
    # load entry points within function, not at module level,
    # to avoid circular dependencies
    # (user would be unable to import networks in other packages
    # that subclass vak.network.AbstractVakNetwork
    # since the module in the other package would need to `import vak`)
    NETWORKS = network._load()
    NETWORK_NAMES = NETWORKS.keys()
    try:
        networks = [network_name for network_name in
                    config['TRAIN']['networks'].split(',')]
        for network_name in networks:
            if network_name not in NETWORK_NAMES:
                raise TypeError('Neural network {} not found when importing installed networks.'
                                .format(network))
        config_dict['networks'] = networks
    except NoOptionError:
        raise KeyError("'networks' option not found in [TRAIN] section of config.ini file. "
                       "Please add this option as a comma-separated list of neural network names, e.g.:\n"
                       "networks = TweetyNet, GRUnet, convnet")

    try:
        config_dict['train_data_dict_path'] = config['TRAIN']['train_data_path']
    except NoOptionError:
        raise KeyError("'train_data_path' option not found in [TRAIN] section of config.ini file. "
                       "Please add this option.")

    if config.has_option('TRAIN', 'train_set_durs'):
        config_dict['train_set_durs'] = [int(element)
                                         for element in
                                         config['TRAIN']['train_set_durs'].split(',')]

    if config.has_option('TRAIN', 'replicates'):
        config_dict['num_replicates'] = int(config['TRAIN']['replicates'])

    if config.has_option('TRAIN', 'val_data_path'):
        config_dict['val_data_dict_path'] = config['TRAIN']['val_data_path']

    if config.has_option('TRAIN', 'test_data_path'):
        config_dict['test_data_dict_path'] = config['TRAIN']['test_data_path']

    if config.has_option('TRAIN', 'val_error_step'):
        config_dict['val_error_step'] = int(config['TRAIN']['val_error_step'])

    if config.has_option('TRAIN', 'checkpoint_step'):
        config_dict['checkpoint_step'] = int(config['TRAIN']['checkpoint_step'])

    if config.has_option('TRAIN', 'save_only_single_checkpoint_file'):
        config_dict['save_only_single_checkpoint_file'] = config.getboolean(
            'TRAIN','save_only_single_checkpoint_file'
        )

    if config.has_option('TRAIN', 'num_epochs'):
        config_dict['num_epochs'] = int(config['TRAIN']['num_epochs'])

    if config.has_option('TRAIN', 'patience'):
        patience = config['TRAIN']['patience']
        try:
            patience = int(patience)
        except ValueError:
            if patience == 'None':
                patience = None
            else:
                raise TypeError('patience must be an int or None, but'
                                'is {} and parsed as type {}'
                                .format(patience, type(patience)))
        config_dict['patience'] = patience

    if config.has_option('TRAIN', 'normalize_spectrograms'):
        config_dict['normalize_spectrograms'] = config.getboolean(
            'TRAIN', 'normalize_spectrograms'
        )

    if config.has_option('TRAIN', 'use_train_subsets_from_previous_run'):
        config_dict['use_train_subsets_from_previous_run'] = config.getboolean(
            'TRAIN', 'use_train_subsets_from_previous_run')
        if config_dict['use_train_subsets_from_previous_run']:
            try:
                config_dict['previous_run_path'] = config['TRAIN']['previous_run_path']
            except KeyError:
                raise KeyError('In config.file {}, '
                               'use_train_subsets_from_previous_run = Yes, but '
                               'no previous_run_path option was found.'
                               'Please add previous_run_path to config file.'
                               .format(config_file))
        else:
            if config.has_option('TRAIN', 'previous_run_path'):
                raise ValueError('In config.file {}, '
                                 'use_train_subsets_from_previous_run = No, but '
                                 'previous_run_path option was specified as {}.\n'
                                 'Please fix argument or remove/comment out '
                                 'previous_run_path.'
                                 .format(config_file,
                                         config['TRAIN']['previous_run_path'])
                                 )

    if config.has_option('TRAIN', 'save_transformed_data'):
        config_dict['save_transformed_data'] = config.getboolean(
            'TRAIN', 'save_transformed_data')

    return TrainConfig(**config_dict)



import numpy as np


def levenshtein(source, target):
    """levenshtein distance
    returns number of deletions, insertions, or substitutions
    required to convert source string into target string.

    Parameters
    ----------
    source : str
        in this context, predicted labels for songbird syllables
    target : str
        in this context, ground truth labels for songbird syllables

    Returns
    -------
    distance : int
        number of deletions, insertions, or substitutions
        required to convert source into target.

    from https://en.wikibooks.org/wiki/Algorithm_Implementation/Strings/Levenshtein_distance#Python
    """

    if len(source) < len(target):
        return levenshtein(target, source)

    # So now we have len(source) >= len(target).
    if len(target) == 0:
        return len(source)

    # We call tuple() to force strings to be used as sequences
    # ('c', 'a', 't', 's') - numpy uses them as values by default.
    source = np.array(tuple(source))
    target = np.array(tuple(target))

    # We use a dynamic programming algorithm, but with the
    # added optimization that we only need the last two rows
    # of the matrix.
    previous_row = np.arange(target.size + 1)
    for s in source:
        # Insertion (target grows longer than source):
        current_row = previous_row + 1

        # Substitution or matching:
        # Target and source items are aligned, and either
        # are different (cost of 1), or are the same (cost of 0).
        current_row[1:] = np.minimum(
                current_row[1:],
                np.add(previous_row[:-1], target != s))

        # Deletion (target grows shorter than source):
        current_row[1:] = np.minimum(
                current_row[1:],
                current_row[0:-1] + 1)

        previous_row = current_row

    return previous_row[-1]


def syllable_error_rate(true, pred):
    """syllable error rate: word error rate, but with songbird syllables
    Levenshtein/edit distance normalized by length of true sequence

    Parameters
    ----------
    true : str
        ground truth labels for a series of songbird syllables
    pred : str
        predicted labels for a series of songbird syllables

    Returns
    -------
    Levenshtein distance / len(true)
    """

    if type(true) != str or type(pred) != str:
        raise TypeError('Both `true` and `pred` must be of type `str')

    return levenshtein(pred, true) / len(true)



"""sub-package with function that parses config.ini file and returns config object"""
from .parse import parse_config
from .data import parse_data_config
from .output import parse_output_config
from .predict import parse_predict_config
from .spectrogram import parse_spect_config
from .train import parse_train_config




import os
from configparser import ConfigParser
from configparser import NoSectionError, MissingSectionHeaderError, ParsingError,\
    DuplicateOptionError, DuplicateSectionError

import attr
from attr.validators import instance_of, optional

from .data import parse_data_config, DataConfig
from .spectrogram import parse_spect_config, SpectConfig
from .train import parse_train_config, TrainConfig
from .output import parse_output_config, OutputConfig
from .predict import parse_predict_config, PredictConfig
from .. import network


def _get_nets_config(config_obj, networks):
    """helper function to get configuration for only networks that user specified
    in a specific section of config.ini file, e.g. in the TRAIN section

    Parameters
    ----------
    config_obj : configparser.ConfigParser
        instance of ConfigParser with config.ini file already read into it
        that has sections representing configurations for networks
    networks : list
        of str, i.e. names of networks specified by a section
        (such as TRAIN or PREDICT) that should each have corresponding sections
        specifying their configuration: hyperparameters such as learning
        rate, number of time steps, etc.

    Returns
    -------
    networks_dict : dict
        where each key is the name of a network and the corresponding value is
        another dict containing key-value pairs of configuration options and the
        value specified for that option
    """
    # load entry points within function, not at module level,
    # to avoid circular dependencies
    # (user would be unable to import networks in other packages
    # that subclass vak.network.AbstractVakNetwork
    # since the module in the other package would need to `import vak`)
    NETWORKS = network._load()
    sections = config_obj.sections()
    networks_dict = {}
    for network_name in networks:
        if network_name not in sections:
            raise NoSectionError('No section found specifying parameters for network {}'
                                 .format(network_name))
        network_option_names = set(config_obj[network_name].keys())
        config_field_names = set(NETWORKS[network_name].Config._fields)
        # if some options in this network's section are not found in the Config tuple
        # that is a class attribute for that network, raise an error because we don't
        # know what to do with that option
        if not network_option_names.issubset(config_field_names):
            unknown_options = network_option_names - config_field_names
            raise ValueError('The following option(s) in section for network {} are '
                             'not found in the Config for that network: {}.\n'
                             'Valid options are: {}'
                             .format(network_name, unknown_options, config_field_names))

        options = {}
        # do type conversion using the networks' Config typed namedtuple
        # for the rest of the options
        for option, value in config_obj[network_name].items():
            option_type = NETWORKS[network_name].Config._field_types[option]
            try:
                options[option] = option_type(value)
            except ValueError:
                raise ValueError('Could not cast value {} for option {} in {} section '
                                 ' to specified type {}.'
                                 .format(value, option, network_name, option_type))
        networks_dict[network_name] = NETWORKS[network_name].Config(**options)
        return networks_dict


@attr.s
class Config:
    """class to represent config.ini file

    Attributes
    ----------
    data : vak.config.data.DataConfig
        represents [DATA] section of config.ini file
    spect_params : vak.config.spectrogram.SpectConfig
        represents [SPECTROGRAM] section of config.ini file
    train : vak.config.train.TrainConfig
        represents [TRAIN] section of config.ini file
    predict : vak.config.predict.PredictConfig
        represents [PREDICT] section of config.ini file.
    output : vak.config.output.OutputConfig
        represents [OUTPUT] section of config.ini file
    networks : dict
        contains neural network configuration sections of config.ini file.
        These will vary depending on which networks the user specifies.
    """
    data = attr.ib(validator=optional(instance_of(DataConfig)), default=None)
    spect_params = attr.ib(validator=optional(instance_of(SpectConfig)), default=None)
    train = attr.ib(validator=optional(instance_of(TrainConfig)), default=None)
    predict = attr.ib(validator=optional(instance_of(PredictConfig)), default=None)
    output = attr.ib(validator=optional(instance_of(OutputConfig)), default=None)
    networks = attr.ib(validator=optional(instance_of(dict)), default=None)


def parse_config(config_file):
    """parse a config.ini file

    Parameters
    ----------
    config_file : str
        path to config.ini file

    Returns
    -------
    config : vak.config.parse.Config
        instance of Config class, whose attributes correspond to
        sections in a config.ini file.
    """
    # check config_file exists,
    # because if it doesn't ConfigParser will just return an "empty" instance w/out sections or options
    if not os.path.isfile(config_file):
        raise FileNotFoundError('config file {} is not found'
                                .format(config_file))

    try:
        config_obj = ConfigParser()
        config_obj.read(config_file)
    except (MissingSectionHeaderError, ParsingError, DuplicateOptionError, DuplicateSectionError):
        # try to add some context for users that do not spend their lives thinking about ConfigParser objects
        print(f"Error when opening the following config_file: {config_file}")
        raise
    except:
        # say something different if we can't add very good context
        print(f"Unexpected error when opening the following config_file: {config_file}")
        raise

    config_dict = {}
    if config_obj.has_section('DATA'):
        config_dict['data'] = parse_data_config(config_obj, config_file)

    ### if **not** using spectrograms from .mat files ###
    if config_obj.has_section('SPECTROGRAM'):
        config_dict['spect_params'] = parse_spect_config(config_obj)

    networks = []
    if config_obj.has_section('TRAIN'):
        config_dict['train'] = parse_train_config(config_obj, config_file)
        networks += config_dict['train'].networks

    if config_obj.has_section('PREDICT'):
        config_dict['predict'] = parse_predict_config(config_obj)
        networks += config_dict['predict'].networks

    if networks:
        config_dict['networks'] = _get_nets_config(config_obj, networks)

    if config_obj.has_section('OUTPUT'):
        config_dict['output'] = parse_output_config(config_obj)

    return Config(**config_dict)



"""
evfuncs
Python implementations of functions used with EvTAF and evsonganaly.m
adapated from hybrid-vocal-classifier
https://github.com/NickleDave/hybrid-vocal-classifier
under BSD license
https://github.com/NickleDave/hybrid-vocal-classifier/blob/master/LICENSE
"""

import os

import numpy as np
import scipy
from scipy.io import loadmat


def readrecf(filename):
    """
    reads .rec files output by EvTAF
    """

    rec_dict = {}
    with open(filename, 'r') as recfile:
        line_tmp = ""
        while 1:
            if line_tmp == "":
                line = recfile.readline()
            else:
                line = line_tmp
                line_tmp = ""

            if line == "":  # if End Of File
                break
            elif line == "\n":  # if blank line
                continue
            elif "Catch" in line:
                ind = line.find('=')
                rec_dict['iscatch'] = line[ind + 1:]
            elif "Chans" in line:
                ind = line.find('=')
                rec_dict['num_channels'] = int(line[ind + 1:])
            elif "ADFREQ" in line:
                ind = line.find('=')
                try:
                    rec_dict['sample_freq'] = int(line[ind + 1:])
                except ValueError:
                    rec_dict['sample_freq'] = float(line[ind + 1:])
            elif "Samples" in line:
                ind = line.find('=')
                rec_dict['num_samples'] = int(line[ind + 1:])
            elif "T After" in line:
                ind = line.find('=')
                rec_dict['time_after'] = float(line[ind + 1:])
            elif "T Before" in line:
                ind = line.find('=')
                rec_dict['time before'] = float(line[ind + 1:])
            elif "Output Sound File" in line:
                ind = line.find('=')
                rec_dict['outfile'] = line[ind + 1:]
            elif "Thresholds" in line:
                th_list = []
                while 1:
                    line = recfile.readline()
                    if line == "":
                        break
                    try:
                        th_list.append(float(line))
                    except ValueError:  # because we reached next section
                        line_tmp = line
                        break
                rec_dict['thresholds'] = th_list
                if line == "":
                    break
            elif "Feedback information" in line:
                fb_dict = {}
                while 1:
                    line = recfile.readline()
                    if line == "":
                        break
                    elif line == "\n":
                        continue
                    ind = line.find("msec")
                    time = float(line[:ind - 1])
                    ind = line.find(":")
                    fb_type = line[ind + 2:]
                    fb_dict[time] = fb_type
                rec_dict['feedback_info'] = fb_dict
                if line == "":
                    break
            elif "File created" in line:
                header = [line]
                for counter in range(4):
                    line = recfile.readline()
                    header.append(line)
                rec_dict['header'] = header
    return rec_dict


def load_cbin(filename, channel=0):
    """
    loads .cbin files output by EvTAF.

    arguments
    ---------
    filename : string

    channel : integer
        default is 0

    returns
    -------
    data : numpy array
        1-d vector of 16-bit signed integers

    sample_freq : integer
        sampling frequency in Hz. Typically 32000.
    """

    # .cbin files are big endian, 16 bit signed int, hence dtype=">i2" below
    data = np.fromfile(filename, dtype=">i2")
    recfile = filename[:-5] + '.rec'
    rec_dict = readrecf(recfile)
    data = data[channel::rec_dict['num_channels']]  # step by number of channels
    sample_freq = rec_dict['sample_freq']
    return data, sample_freq


def load_notmat(filename):
    """
    loads .not.mat files created by evsonganaly.m.
    wrapper around scipy.io.loadmat.
    Calls loadmat with squeeze_me=True to remove extra dimensions from arrays
    that loadmat parser sometimes adds.

    Argument
    --------
    filename : string, name of .not.mat file

    Returns
    -------
    notmat_dict : dictionary of variables from .not.mat files
    """

    if ".not.mat" in filename:
        pass
    elif filename[-4:] == "cbin":
        filename += ".not.mat"
    else:
        raise ValueError("Filename should have extension .cbin.not.mat or"
                         " .cbin")

    if not os.path.isfile(filename):
        raise FileNotFoundError
    else:
        return loadmat(filename, squeeze_me=True)


def bandpass_filtfilt(rawsong, samp_freq, freq_cutoffs=None):
    """filter song audio with band pass filter, run through filtfilt
    (zero-phase filter)

    Parameters
    ----------
    rawsong : ndarray
        audio
    samp_freq : int
        sampling frequency
    freq_cutoffs : list
        2 elements long, cutoff frequencies for bandpass filter
        if None, set to [500, 10000]. Default is None.

    Returns
    -------
    filtsong : ndarray
    """

    Nyquist_rate = samp_freq / 2
    if freq_cutoffs is None:
        freq_cutoffs = [500, 10000]
    if rawsong.shape[-1] < 387:
        numtaps = 64
    elif rawsong.shape[-1] < 771:
        numtaps = 128
    elif rawsong.shape[-1] < 1539:
        numtaps = 256
    else:
        numtaps = 512

    cutoffs = np.asarray([freq_cutoffs[0] / Nyquist_rate,
                          freq_cutoffs[1] / Nyquist_rate])
    # code on which this is based, bandpass_filtfilt.m, says it uses Hann(ing)
    # window to design filter, but default for matlab's fir1
    # is actually Hamming
    # note that first parameter for scipy.signal.firwin is filter *length*
    # whereas argument to matlab's fir1 is filter *order*
    # for linear FIR, filter length is filter order + 1
    b = scipy.signal.firwin(numtaps + 1, cutoffs, pass_zero=False)
    a = np.zeros((numtaps+1,))
    a[0] = 1  # make an "all-zero filter"
    padlen = np.max((b.shape[-1] - 1, a.shape[-1] - 1))
    filtsong = scipy.signal.filtfilt(b, a, rawsong, padlen=padlen)
    return filtsong


def smooth_data(rawsong, samp_freq, freq_cutoffs=None, smooth_win=2):
    """filter raw audio and smooth signal
    used to calculate amplitude.

    Parameters
    ----------
    rawsong : 1-d numpy array
        "raw" voltage waveform from microphone
    samp_freq : int
        sampling frequency
    freq_cutoffs: list
        two-element list of integers, [low freq., high freq.]
        bandpass filter applied with this list defining pass band.
        Default is None, in which case bandpass filter is not applied.
    smooth_win : integer
        size of smoothing window in milliseconds. Default is 2.

    Returns
    -------
    smooth : 1-d numpy array
        smoothed waveform

    Applies a bandpass filter with the frequency cutoffs in spect_params,
    then rectifies the signal by squaring, and lastly smooths by taking
    the average within a window of size sm_win.
    This is a very literal translation from the Matlab function SmoothData.m
    by Evren Tumer. Uses the Thomas-Santana algorithm.
    """

    if freq_cutoffs is None:
        # then don't do bandpass_filtfilt
        filtsong = rawsong
    else:
        filtsong = bandpass_filtfilt(rawsong, samp_freq, freq_cutoffs)

    squared_song = np.power(filtsong, 2)
    len = np.round(samp_freq * smooth_win / 1000).astype(int)
    h = np.ones((len,)) / len
    smooth = np.convolve(squared_song, h)
    offset = round((smooth.shape[-1] - filtsong.shape[-1]) / 2)
    smooth = smooth[offset:filtsong.shape[-1] + offset]
    return smooth


"""spectrogram utilities
filters adapted from SciPy cookbook
https://scipy-cookbook.readthedocs.io/items/ButterworthBandpass.html
spectrogram adapted from code by Kyle Kastner and Tim Sainburg
https://github.com/timsainb/python_spectrograms_and_inversion
"""

import numpy as np
from scipy.signal import butter, lfilter
from matplotlib.mlab import specgram


def butter_bandpass(lowcut, highcut, fs, order=5):
    nyq = 0.5 * fs
    low = lowcut / nyq
    high = highcut / nyq
    b, a = butter(order, [low, high], btype='band')
    return b, a


def butter_bandpass_filter(data, lowcut, highcut, fs, order=5):
    b, a = butter_bandpass(lowcut, highcut, fs, order=order)
    y = lfilter(b, a, data)
    return y


def spectrogram(data, samp_freq, fft_size=512, step_size=64, thresh=None, transform_type=None):
    """creates a spectrogram

    Parameters
    ----------
    data : ndarray
        audio signal
    log_transform: bool
        if True, take the log of the spectrogram
    thresh: int
        threshold minimum power for log spectrogram

    Return
    ------
    spec : ndarray

    freqbins : ndarray

    timebins : ndarray
    """

    noverlap = fft_size - step_size

    # below only take [:3] from return of specgram because we don't need the image
    spec, freqbins, timebins = specgram(data, fft_size, samp_freq, noverlap=noverlap)[:3]

    if transform_type:
        if transform_type == 'log_spect':
            spec /= spec.max()  # volume normalize to max 1
            spec = np.log10(spec)  # take log
            if thresh:
                # I know this is weird, maintaining 'legacy' behavior
                spec[spec < -thresh] = -thresh
        elif transform_type == 'log_spect_plus_one':
            spec = np.log10(spec + 1)
            if thresh:
                spec[spec < thresh] = thresh
    else:
        if thresh:
            spec[spec < thresh] = thresh  # set anything less than the threshold as the threshold

    return spec, freqbins, timebins
